#include "editor.h"
#include "bridge.h"
#include "../common/sources/system/cgraphics2d.h"
#include "../common/sources/system/filecache.h"
#include "tools.h"
#include "rectangles.h"
#include <QDebug>
#include <QtXml/qdom.h>
#include <QFile>
#include <QQmlApplicationEngine>

CEditor  editor;
extern QQmlApplicationEngine*  myQmlEngine;
extern QQuickItem*  tileSelGrid;
extern QQuickItem*  wallSelGrid;
extern QQuickItem*  objSelGrid;

//---------------------------------------------------------------------------------------------
CEditor::CEditor()
{
    mUndoStack = new QUndoStack();
    mSelectStart.x = -1;
}

//---------------------------------------------------------------------------------------------
// get the wall side for a position relative to a tile
EWallSide   CEditor::getWallSide(CVec2 localPos)
{
    bool side1 = (localPos.y < localPos.x);
    bool side2 = (localPos.y < TILE_SIZE - 1 - localPos.x);

    if (side1 == true && side2 == true)
        return eWallSideUp;

    if (side1 == true && side2 == false)
        return eWallSideRight;

    if (side1 == false && side2 == true)
        return eWallSideLeft;

    return eWallSideDown;
}

//---------------------------------------------------------------------------------------------
// get the wall side for the mouse position in the map
EWallSide   CEditor::getWallSideAbs(CVec2 pos)
{
    return getWallSide(pos % TILE_SIZE);
}

//---------------------------------------------------------------------------------------------
CRect CEditor::getSelectionRect()
{
    CRect   rect = getSelectionRectPixel();

    rect.tl /= TILE_SIZE;
    rect.br /= TILE_SIZE;
    return rect;
}

//---------------------------------------------------------------------------------------------
CRect CEditor::getSelectionRectPixel()
{
    CRect   rect;

    // handle the cases when we do not draw the rectangle from top/left to bottom/right
    rect.tl.x = MIN(mSelectStart.x, mSelectEnd.x);
    rect.tl.y = MIN(mSelectStart.y, mSelectEnd.y);
    rect.br.x = MAX(mSelectStart.x, mSelectEnd.x);
    rect.br.y = MAX(mSelectStart.y, mSelectEnd.y);
    return rect;
}


//---------------------------------------------------------------------------------------------
// returns the string that tells the position of the cursor in the map
QString CEditor::getPositionString(CVec2 mousePos, ETabIndex tabIndex)
{
    CVec2 tilePos = mousePos / TILE_SIZE;
    static char tempString[256];
    tempString[0] = 0;

    if (tabIndex == eTabTiles)
    {
        sprintf(tempString, "(%d,%d)", tilePos.x, tilePos.y);
    }
    else if (tabIndex == eTabWalls)
    {
        EWallSide   side = getWallSideAbs(mousePos);

        switch (side)
        {
            case eWallSideUp:
                sprintf(tempString, "(%d,%d) up", tilePos.x, tilePos.y);
            break;

            case eWallSideDown:
                sprintf(tempString, "(%d,%d) down", tilePos.x, tilePos.y);
            break;

            case eWallSideLeft:
                sprintf(tempString, "(%d,%d) left", tilePos.x, tilePos.y);
            break;

            case eWallSideRight:
                sprintf(tempString, "(%d,%d) right", tilePos.x, tilePos.y);
            break;

            default:
            break;
        }
    }
    else if (tabIndex == eTabObjects)
    {
        CVec2   objPos = (mousePos % TILE_SIZE) / (TILE_SIZE / 2);
        sprintf(tempString, "(%d,%d) (%d,%d)", tilePos.x, tilePos.y, objPos.x, objPos.y);
    }

    return QString(tempString);
}

//---------------------------------------------------------------------------------------------
void   CEditor::hideSelectionInfos()
{
    bridge.set_tileSelScrollVis(false);
    bridge.set_wallSelScrollVis(false);
    bridge.set_objectSelScrollVis(false);
    bridge.set_rectSelScrollVis(false);
}

//---------------------------------------------------------------------------------------------
void CEditor::displayWall(QImage* image, CVec2 pos, EWallSide side, int angle)
{
    int type = map.getTile(pos)->mWalls[side].getType();
    QString file = wallsDatas[type].image;

    if (file.isEmpty() == false)
    {
        QImage  tileImage = fileCache.getImage(file.toLocal8Bit().constData());
        CVec2   pos2 = pos * TILE_SIZE;
        graph2D.drawImage(image, pos2, tileImage, angle);
    }
}

//---------------------------------------------------------------------------------------------
QImage CEditor::displayMap(CMap& map, CVec2 mousePos, ETabIndex tabIndex)
{
    QImage image(map.mSize.x * TILE_SIZE,
                 map.mSize.y * TILE_SIZE,
                 QImage::Format_RGBA8888);

    image.fill(QColor(0, 0, 0));

    if (map.isLoaded() == true)
    {
        // draw tiles
        for (int y = 0; y < map.mSize.y; ++y)
            for (int x = 0; x < map.mSize.x; ++x)
            {
                CVec2   pos(x, y);
                CVec2   pos2 = pos * TILE_SIZE;

                int type = map.getTile(pos)->mType;
                QString file = tilesDatas[type].image;

                if (file.isEmpty() == false)
                {
                    QImage  tileImage = fileCache.getImage(file.toLocal8Bit().constData());
                    graph2D.drawImage(&image, pos2, tileImage);
                }
            }

        // draw walls
        for (int y = 0; y < map.mSize.y; ++y)
            for (int x = 0; x < map.mSize.x; ++x)
            {
                CVec2   pos(x, y);

                displayWall(&image, pos, eWallSideUp, 0);
                displayWall(&image, pos, eWallSideDown, 180);
                displayWall(&image, pos, eWallSideLeft, -90);
                displayWall(&image, pos, eWallSideRight, 90);
            }
    }

    rectangles.drawSelection(&image, tabIndex);
    rectangles.drawCursor(&image, mousePos, tabIndex);
    rectangles.drawPasteSelection(&image);

    return image;
}

//---------------------------------------------------------------------------------------------
void    CEditor::useTool()
{
    switch (bridge.mTool)
    {
    case eQMLToolDraw:
        toolDraw();
        break;

    case eQMLToolSelect:
        toolSelect();
        break;

    case eQMLToolPaste:
        toolPaste();
        break;
    }
}

//---------------------------------------------------------------------------------------------
void    CEditor::toolDraw()
{
    if (bridge.rightPressed == true || bridge.leftPressed == true)
    {
        if (bridge.mTabIndex == eTabTiles)
        {
            // tiles
            drawTileTool*   tool = new drawTileTool();
            int type = 0;

            if (bridge.rightPressed == true)
                type = 0;
            else if (bridge.leftPressed == true)
                type = bridge.get_tileType();

            if (tool->init(bridge.mMousePos, type) == true)
            {
                unselect();
                mUndoStack->push(tool);
            }
            else
                delete tool;
        }
        else if (bridge.mTabIndex == eTabWalls)
        {
            // walls
            drawWallTool*   tool = new drawWallTool();
            int type = 0;

            if (bridge.rightPressed == true)
                type = 0;
            else if (bridge.leftPressed == true)
                type = bridge.get_wallType();

            if (tool->init(bridge.mMousePos, type) == true)
            {
                unselect();
                mUndoStack->push(tool);
            }
            else
                delete tool;
        }
    }
}

//---------------------------------------------------------------------------------------------
// Did we select a single wall ?
bool   CEditor::isSelectSingleWall()
{
    CVec2   startPos = mSelectStart / TILE_SIZE;
    CVec2   endPos   = mSelectEnd   / TILE_SIZE;
    EWallSide   startSide = getWallSideAbs(mSelectStart);
    EWallSide   endSide   = getWallSideAbs(mSelectEnd);

    return (startPos == endPos && startSide == endSide);
}

//---------------------------------------------------------------------------------------------
// Did we select a single object ?
bool   CEditor::isSelectSingleObject()
{
    CVec2   startPos = mSelectStart / (TILE_SIZE / 2);
    CVec2   endPos   = mSelectEnd   / (TILE_SIZE / 2);

    return (startPos == endPos);
}

//---------------------------------------------------------------------------------------------
void        CEditor::copyData(CRect rect)
{
    CVec2   size = (rect.br / TILE_SIZE) - (rect.tl / TILE_SIZE) + CVec2(1, 1);
    mClipboard.copyFromMap(rect.tl, size, bridge.mTabIndex);
}

//---------------------------------------------------------------------------------------------
void        CEditor::toolCopy()
{
    if (mSelectStart.x != -1)
    {
        CRect   rect = getSelectionRectPixel();
        copyData(rect);
        unselect();
    }
}

//---------------------------------------------------------------------------------------------
void        CEditor::toolCut()
{
    if (mSelectStart.x != -1)
    {
        cutTool*   tool = new cutTool();
        CRect   rect = getSelectionRectPixel();
        copyData(rect);

        if (tool->init(mSelectStart, bridge.mTabIndex) == true)
            mUndoStack->push(tool);
        else
            delete tool;
    }
    unselect();
}

//---------------------------------------------------------------------------------------------
void        CEditor::toolPaste()
{
    if (bridge.leftPressing == true)
    {
        unselect();
        pasteTool*   tool = new pasteTool();

        if (tool->init(bridge.mMousePos, bridge.mTabIndex) == true)
            mUndoStack->push(tool);
        else
            delete tool;
    }
}

//---------------------------------------------------------------------------------------------
void    CEditor::unselect()
{
    mSelectStart.x = -1;
    bridge.updateQMLImage();
}

//---------------------------------------------------------------------------------------------
// destroy the additionnal parameters in the selection info groupBox
void    CEditor::deleteParamItems(std::vector<QQuickItem*>* list)
{
    for (size_t i = 0; i < list->size(); ++i)
        delete (*list)[i];
    list->clear();
}

//---------------------------------------------------------------------------------------------
// creates the additionnal parameters in the selection info groupBox
void    CEditor::addWallParamsItems(uint8_t type)
{
    CVec2       pos = mSelectStart / TILE_SIZE;
    EWallSide   side = getWallSideAbs(mSelectStart);
    CWall*      wall = &map.getTile(pos)->mWalls[side];

    std::vector<CParamType>& sourceList = map.mWallsParams[type];

    for (size_t i = 0; i < sourceList.size(); ++i)
    {
        CParam* param = wall->mParams[i];

        if (sourceList[i].mType == eParamOrnate)
        {
            CParamOrnate*   par = (CParamOrnate*)param;
            addLabel(wallSelGrid, sourceList[i].mName + ":", &mWallParamItems);
            addComboBox(wallSelGrid, bridge.ornatesList, &mWallParamItems, i, par->mValue);
        }
    }
}

//---------------------------------------------------------------------------------------------
void    CEditor::toolSelect()
{
    if (bridge.leftPressed == true)
    {
        mSelectEnd = bridge.mMousePos;

        if (bridge.leftPressing == true)
            mSelectStart = mSelectEnd;

        // display infos in the selection groupbox
        CRect   rectPixel = getSelectionRectPixel();
        CRect   rect = CRect(rectPixel.tl / TILE_SIZE, rectPixel.br / TILE_SIZE);
        QString positionText = getPositionString(rectPixel.tl, bridge.mTabIndex);
        bridge.set_selPosition(positionText);
        CVec2 selSize = rect.br - rect.tl + CVec2(1, 1);
        static char sizeText[256];
        sprintf(sizeText, "(%d,%d)", selSize.x, selSize.y);
        bridge.set_selSize(QString(sizeText));

        switch (bridge.mTabIndex)
        {
        case eTabTiles:
            if (selSize.x == 1 && selSize.y == 1)
                bridge.set_selTileType(map.getTile(rect.tl)->mType);
            break;

        case eTabWalls:
            if (isSelectSingleWall() == true)
            {
                EWallSide   side = getWallSideAbs(mSelectStart);
                uint8_t     type = map.getTile(rect.tl)->mWalls[side].getType();
                bridge.set_selWallType(type);

                deleteParamItems(&mWallParamItems);
                addWallParamsItems(type);
            }
            break;

        default:
            break;
        }

        bridge.updateQMLImage();
    }
}

//---------------------------------------------------------------------------------------------
void    CEditor::readTilesDB(QString fileName)
{
    QDomDocument doc;

    QFile f(fileName);
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomElement tile = root.firstChild().toElement();

    while(!tile.isNull())
    {
        if (tile.tagName() == "tile")
        {
            CTileData   newTile;

            newTile.name = tile.attribute("name");

            QDomElement tileInfo = tile.firstChild().toElement();

            while(!tileInfo.isNull())
            {
                if (tileInfo.tagName() == "image")
                    newTile.image = QString("tiles/") + tileInfo.text();

                tileInfo = tileInfo.nextSibling().toElement();
            }
            tilesDatas.push_back(newTile);
        }
        tile = tile.nextSibling().toElement();
    }
}

//---------------------------------------------------------------------------------------------
void    CEditor::readWallsDB(QString fileName)
{
    QDomDocument doc;

    QFile f(fileName);
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomElement wall = root.firstChild().toElement();

    while(!wall.isNull())
    {
        std::vector<CParamType> paramsTypesList;

        if (wall.tagName() == "wall")
        {
            CWallData   newWall;

            newWall.name = wall.attribute("name");

            QDomElement wallInfo = wall.firstChild().toElement();

            while(!wallInfo.isNull())
            {
                if (wallInfo.tagName() == "image")
                {
                    newWall.image = QString("tiles/") + wallInfo.text();
                }
                else if (wallInfo.tagName() == "param")
                {
                    QString paramType = wallInfo.attribute("type");

                    if (paramType == "ornate")
                    {
                        CParamType  newParamType;
                        newParamType.mType = eParamOrnate;
                        newParamType.mName = wallInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                }

                wallInfo = wallInfo.nextSibling().toElement();
            }
            wallsDatas.push_back(newWall);
            map.mWallsParams.push_back(paramsTypesList);
        }
        wall = wall.nextSibling().toElement();
    }
}

//---------------------------------------------------------------------------------------------
void    CEditor::readOrnatesDB(QString fileName)
{
    QDomDocument doc;

    QFile f(fileName);
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomElement ornate = root.firstChild().toElement();

    int i = 0;
    while(!ornate.isNull())
    {
        if (ornate.tagName() == "ornate")
        {
            bridge.ornatesList.append(QString("%1 - %2").arg(QString::number(i), ornate.attribute("name")));
            i++;
        }
        ornate = ornate.nextSibling().toElement();
    }
}

//---------------------------------------------------------------------------------------------
void    CEditor::initLists()
{
    readTilesDB("databases/tiles.xml");

    for (size_t i = 0; i < tilesDatas.size(); ++i)
        bridge.tilesList.append(QString("%1 - %2").arg(QString::number(i), tilesDatas[i].name));

    readWallsDB("databases/walls.xml");

    for (size_t i = 0; i < wallsDatas.size(); ++i)
        bridge.wallsList.append(QString("%1 - %2").arg(QString::number(i), wallsDatas[i].name));

    readOrnatesDB("databases/ornates.xml");
}

//---------------------------------------------------------------------------------------------
QQuickItem* CEditor::loadQml(QString file, QQuickItem* parent)
{
    QQmlComponent   component(myQmlEngine, QUrl(file));
    QQuickItem*     object = qobject_cast<QQuickItem*>(component.create());
    QQmlEngine::setObjectOwnership(object, QQmlEngine::CppOwnership);
    object->setParent(parent);
    object->setParentItem(parent);
    return object;
}

//---------------------------------------------------------------------------------------------
void CEditor::addLabel(QQuickItem* parent, QString text, std::vector<QQuickItem *> *list)
{
    QQuickItem* object = loadQml("qrc:/ParamLabel.qml", parent);
    object->setProperty("text", QVariant(text));
    list->push_back(object);
}

//---------------------------------------------------------------------------------------------
void CEditor::addComboBox(QQuickItem* parent, QStringList& values, std::vector<QQuickItem*>* list, int id, int index)
{
    QQuickItem* object = loadQml("qrc:/ParamComboBox.qml", parent);
    object->setProperty("model", QVariant(values));
    object->setProperty("identifier", QVariant(id));
    object->setProperty("currentIndex", QVariant(index));
    list->push_back(object);
}
